︠eeff1cb3-9f85-44b1-8039-ee801ede8610i︠
%html
<h1>Taylor Polynomials</h1>
<p>Since Taylor polynomials concentrate their accuracy around a point (making sure the function and polynomial have the same derivatives up to the $n$th derivative at that point), a Taylor polynomial may differ from a function significantly away from the point.</p>
<p>In the example below, we calculate and plot the Taylor polynomial $\hat{f}$ approximating the function $f(x)=1/x$ around the point $x=1$.&nbsp; We also print the error in the approximation at $x=3$, which is $|f(3)-\hat{f}(3)|$.&nbsp; Higher order Taylor polynomials usually give better approximations farther from the point, but in this case, as we increase the order of the approximation, the error gets much, much worse.&nbsp; Check this out by moving the slider to change the order of the approximation.</p>

︡aad96827-94b7-48f9-a1bb-e7afc97c408e︡{"html": "<h1>Taylor Polynomials</h1>\r\n<p>Since Taylor polynomials concentrate their accuracy around a point (making sure the function and polynomial have the same derivatives up to the $n$th derivative at that point), a Taylor polynomial may differ from a function significantly away from the point.</p>\r\n<p>In the example below, we calculate and plot the Taylor polynomial $\\hat{f}$ approximating the function $f(x)=1/x$ around the point $x=1$.&nbsp; We also print the error in the approximation at $x=3$, which is $|f(3)-\\hat{f}(3)|$.&nbsp; Higher order Taylor polynomials usually give better approximations farther from the point, but in this case, as we increase the order of the approximation, the error gets much, much worse.&nbsp; Check this out by moving the slider to change the order of the approximation.</p>"}︡
︠fe1f5ebf-fa48-4fd7-b206-7733f45b50f5︠
x0  = 1
f(x)= 1/x
p   = plot(f,-1,5, thickness=2)
dot = point((x0,f(x=x0)),pointsize=80,color='red')
@interact
def _(order=(1..12)):
  taylor = f.taylor(x,x0,order)
  taylor_plot = plot(taylor,-1, 5, color='green', thickness=2)
  error = abs(f(3)-taylor(3))
  html('$f(x)=%s$'%latex(f(x)))
  html('$\hat{f}(x;%s)=%s+\mathcal{O}(x^{%s})$'%(x0,latex(taylor(x)),order+1))
  html('|$f(3)-\hat{f}(3)|= %s$'%(error.n())) 
  show(dot + p + taylor_plot+line([(3,f(3)),(3,taylor(3))],color='purple',linestyle=':'), ymin=-50,ymax=50)
︡92bd634c-1f40-4ef8-b313-18168a504ca4︡︡
︠922dd3ec-89ee-4d97-bf56-1541ed72833ai︠
%html
<h1>Lagrange Polynomials</h1>
<p>Instead of concentrating our accuracy around a specific point, let's make sure that our approximation function is exact at a number of specified points.&nbsp; The Lagrange polynomial for $n+1$ data points is the unique $n$th degree polynomial that passes through those points.&nbsp; We define the polynomial by taking linear combinations of functions so that we have the right values at the points we specify.</p>

︡b9258f4a-3a32-41d3-9588-89497ed7d19b︡{"html": "<h1>Lagrange Polynomials</h1>\r\n<p>Instead of concentrating our accuracy around a specific point, let's make sure that our approximation function is exact at a number of specified points.&nbsp; The Lagrange polynomial for $n+1$ data points is the unique $n$th degree polynomial that passes through those points.&nbsp; We define the polynomial by taking linear combinations of functions so that we have the right values at the points we specify.</p>"}︡
︠70113a44-f75e-4975-aea8-cf985589ce64︠
pts=[(i,1/i) for i in [1..5]]
html.table(pts)
︡8de1f8b5-571c-4cf7-b2e9-75b73b1eed8b︡{"html": "\n<div class=\"notruncate\">\n<table class=\"table_form\">\n<tbody>\n<tr class =\"row-a\">\n<td><span class=\"math\">1</span></td>\n<td><span class=\"math\">1</span></td>\n</tr>\n<tr class =\"row-b\">\n<td><span class=\"math\">2</span></td>\n<td><span class=\"math\">\\frac{1}{2}</span></td>\n</tr>\n<tr class =\"row-a\">\n<td><span class=\"math\">3</span></td>\n<td><span class=\"math\">\\frac{1}{3}</span></td>\n</tr>\n<tr class =\"row-b\">\n<td><span class=\"math\">4</span></td>\n<td><span class=\"math\">\\frac{1}{4}</span></td>\n</tr>\n<tr class =\"row-a\">\n<td><span class=\"math\">5</span></td>\n<td><span class=\"math\">\\frac{1}{5}</span></td>\n</tr>\n</tbody>\n</table>\n</div>\n"}︡
︠766e0bcb-9f96-42e1-ab51-ea60618b1424︠
#pts=[(0,0),(1,1),(2,3)]
︡3fc8631f-543f-43d3-a448-3504ba10a02e︡︡
︠08f5a89c-193f-427b-8178-0c4851096c80︠
def switch_poly(pts,point_to_delete):
    xvalues=[p[0] for p in pts]
    deleted_point=xvalues.pop(point_to_delete)
    numerator=prod((x-pt) for pt in xvalues)
    denominator=numerator(x=deleted_point)
    return numerator/denominator
︡a08179a2-77b5-4528-ab7b-c4a3647670f4︡︡
︠8a2647e4-0100-4155-aa65-749e1ac1900fi︠
%html
<p>For each $x$-coordinate, there is a polynomial that is 1 at that $x$-coordinate and zero at all of the others.&nbsp; Here is the polynomial that is zero at $x_0,x_1,x_3,x_4$ and 1 at $x_2$.</p>

︡0a6020fb-3cba-4bb0-bf8e-8561ec3d435e︡{"html": "<p>For each $x$-coordinate, there is a polynomial that is 1 at that $x$-coordinate and zero at all of the others.&nbsp; Here is the polynomial that is zero at $x_0,x_1,x_3,x_4$ and 1 at $x_2$.</p>"}︡
︠6e387ef7-6d45-4e87-997a-a00351dc70ce︠
plot(switch_poly(pts,2),(x,0.8,5.2),gridlines='major')
︡3467740c-b4cf-4e15-baa1-e764134d7b35︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠7b942148-d186-4bcd-8343-8f0edcb3fab3i︠
%html
<p>Here is a plot of all of the polynomials, one for each $x_i$.</p>

︡b55612f2-2590-4723-93bc-446711f57da4︡{"html": "<p>Here is a plot of all of the polynomials, one for each $x_i$.</p>"}︡
︠e9c42d2c-1ca4-4a17-b119-c026a1aef0b6︠
point_plot=points([(p[0],1) for p in pts],color='red',pointsize=30)
curve_plot=plot([switch_poly(pts,i) for i in range(len(pts))],(x,0.8,5.2),fill=True)
point_plot+curve_plot
︡361bdfdf-7138-415d-8036-72e97f8ab409︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠b083246c-0628-4c36-8f65-c197acdb0872i︠
%html
<p>The Lagrange polynomial is just a linear combination of each of switching polynomials.&nbsp; This guarantees that we have the correct values at the points $x_0$, $x_1$, $x_2$, $x_3$, and $x_4$.</p>

︡9399d0c1-4446-4367-94ad-4a52a30d76e4︡{"html": "<p>The Lagrange polynomial is just a linear combination of each of switching polynomials.&nbsp; This guarantees that we have the correct values at the points $x_0$, $x_1$, $x_2$, $x_3$, and $x_4$.</p>"}︡
︠09ef6f3f-940b-43e2-af66-b30289afe2bd︠
def lagrange_poly(pts):
    return sum(switch_poly(pts,i)*pts[i][1] for i in range(len(pts)))
︡278847cd-da49-4165-8d83-aa0daaac6529︡︡
︠13e2cec5-3427-45cb-b15e-6e30ad7c9a09︠
p(x)=lagrange_poly(pts)
p(x)
︡34cd8ea9-134c-4f64-80d5-f3ba95afdb57︡{"stdout": "1/24*(x - 5)*(x - 4)*(x - 3)*(x - 2) - 1/12*(x - 5)*(x - 4)*(x - 3)*(x - 1) + 1/12*(x - 5)*(x - 4)*(x - 2)*(x - 1) - 1/24*(x - 5)*(x - 3)*(x - 2)*(x - 1) + 1/120*(x - 4)*(x - 3)*(x - 2)*(x - 1)"}︡
︠a1f8166e-9193-4587-a017-e817e4fab036︠
p(x).expand()
︡9af79406-851b-40f7-be67-ecf04df50484︡{"stdout": "1/120*x^4 - 1/8*x^3 + 17/24*x^2 - 15/8*x + 137/60"}︡
︠723ca099-ae98-43ab-8539-3a514b600b34i︠
%html
<p>Sage also has a builtin function to generate Lagrange polynomials as part of its polynomial objects.</p>

︡4c8f9a10-1f7b-443b-bc1b-24fdbc5b7aa1︡{"html": "<p>Sage also has a builtin function to generate Lagrange polynomials as part of its polynomial objects.</p>"}︡
︠3534baa8-4263-4ca6-9cf5-054accd1c347︠
R = PolynomialRing(QQ, 'x')
L = R.lagrange_polynomial(pts)
L
︡ae032478-0cfa-47d6-9eb6-531551629968︡{"stdout": "1/120*x^4 - 1/8*x^3 + 17/24*x^2 - 15/8*x + 137/60"}︡
︠2a3c6418-9c02-4b1e-acbf-cdd598480113i︠
%html
<p>Sage automatically evaluates a Polynomial object using a technique that is at least as efficient as Horner's form (if there are lots of zero coefficients, then it is more efficient by constructing a graph; type "sage.rings.polynomial.polynomial_compiled.CompiledPolynomialFunction??" to see the details and code).&nbsp; Here we evaluate the polynomial at "$x$", and you can see that it looks like Horner's form is being used.</p>

︡d7c27faa-b120-49c2-8258-88787bd6fa69︡{"html": "<p>Sage automatically evaluates a Polynomial object using a technique that is at least as efficient as Horner's form (if there are lots of zero coefficients, then it is more efficient by constructing a graph; type \"sage.rings.polynomial.polynomial_compiled.CompiledPolynomialFunction??\" to see the details and code).&nbsp; Here we evaluate the polynomial at \"$x$\", and you can see that it looks like Horner's form is being used.</p>"}︡
︠02792292-210f-47f1-9a9c-20ae54fb142f︠
L(x)
︡4b70b681-f1d8-4a11-b0de-b8aa74e7f9b1︡{"stdout": "1/120*(((x - 15)*x + 85)*x - 225)*x + 137/60"}︡
︠f59df5cc-0681-4ff8-9660-011b36dbd847i︠
%html
<p>Let's plot this approximation to see how good it is.</p>

︡e2881e16-d079-4a26-97fa-0ba8a0936857︡{"html": "<p>Let's plot this approximation to see how good it is.</p>"}︡
︠2a7c0776-2222-461a-a794-661c70a3bc52︠
x0  = 1
f(x)= 1/x
p   = plot(f,-1,5, thickness=2)
R = PolynomialRing(QQ, 'x')
@interact
def _(x_values=input_box([1,2,3,4,5])):
    pts=[(QQ(i),f(i)) for i in x_values]
    dots = point(pts,pointsize=80,color='red')
    L = R.lagrange_polynomial(pts)
    L_plot = plot(L,(x,-1, 5), color='green', thickness=2)
    error = abs(f(3)-L(x=3))
    html('$f(x)=%s$'%latex(f(x)))
    html('$L(x)=%s=%s$'%(latex(L(x=x)),latex(L(x=x).expand())))
    html('|$f(3)-L(x=3)|= %s$'%(latex(error.n()))) 
    show(dots + p + L_plot+line([(3,f(3)),(3,L(x=3))],color='purple',linestyle=':'), ymin=-10,ymax=10)
︡e1617fab-6b0a-4b99-915e-e93d7b89c13e︡︡
︠703e160f-547b-4067-9fed-4e6f0145c923i︠
%html
<p>These approximations can be pretty bad for some functions.&nbsp; A famous example is the Runge function $$\frac{1}{1+25x^2}$$</p>

︡e1c6b096-dcc6-4c9a-b240-5d018249dc9f︡{"html": "<p>These approximations can be pretty bad for some functions.&nbsp; A famous example is the Runge function $$\\frac{1}{1+25x^2}$$</p>"}︡
︠94ef33af-275e-4cda-b6a1-0e53f03d9db2︠
plot(1/(1+25*x^2), (x,-1,1))
︡5ddb3193-70a4-425d-a476-bcabf24b1bd0︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠537cde40-00fe-4986-8aef-14ebe18880e0︠
R = PolynomialRing(QQ, 'x')
︡08b3678f-92f7-4b14-abfa-26d46a6ea426︡︡
︠36635093-4b37-4074-abf2-a4da4b5adf17︠
f(x)=1/(1+25*x^2)
R.lagrange_polynomial([(i,QQ(f(i))) for i in [0,0.1,0.2,0.3]])
︡5f693e44-c3b6-4463-9462-e1d9e387eadc︡{"stdout": "34.6153846153846*x^3 - 15.3846153846154*x^2 - 0.807692307692308*x + 1.00000000000000"}︡
︠aad82f2b-5523-4047-9afa-20206146f0ef︠
x0  = 1
f(x)= 1/(1+25*x^2)
x_bounds=(-1,1)
p   = plot(f,x_bounds, thickness=2)
R = PolynomialRing(QQ, 'x')
@interact
def _(x_values=input_box([-1,..,1,step=1/20])):
    pts=[(i,QQ(f(i))) for i in x_values]
    dots = point(pts,pointsize=40,color='red')
    L = R.lagrange_polynomial(pts)
    L_plot = plot(L(x),x_bounds, color='green', thickness=2)
    error = abs(f(.995)-L(.995))
    html('$f(x)=%s$'%latex(f(x)))
    html('$L(x)=%s$'%(latex(L(x).expand())))
    html('|$f(.995)-L(.995)|= %s$'%(latex(error.n()))) 
    show(dots + p + L_plot, ymin=-10,ymax=10)
︡36a3be3c-afb6-46fe-9487-823545b3e67f︡︡
︠a1ca8fa1-197e-414d-9add-475ed2ef8f78i︠
%html
<p>It's actually much better to take points that are projections of points evenly spaced around the unit circle (the roots of the "Chebyshev" polynomial).&nbsp; Try putting in "[cos(i/20*pi.n()) for i in [1,3,..,19]]" for the x_values above.</p>

︡036e71be-387d-4b55-82f5-78563fe01965︡{"html": "<p>It's actually much better to take points that are projections of points evenly spaced around the unit circle (the roots of the \"Chebyshev\" polynomial).&nbsp; Try putting in \"[cos(i/20*pi.n()) for i in [1,3,..,19]]\" for the x_values above.</p>"}︡
︠41fe4769-20d2-4f0d-9211-ccd9fbd9b76f︠
[cos(i/20*pi.n()) for i in [1,3,..,19]]
︡0488eca1-5e78-49e6-9b26-aaa75eaebdd9︡{"stdout": "[0.987688340595138, 0.891006524188368, 0.707106781186548, 0.453990499739547, 0.156434465040231, -0.156434465040231, -0.453990499739547, -0.707106781186547, -0.891006524188368, -0.987688340595138]"}︡
︠e2214e7f-4b96-47fa-b2bb-e6a906f29bebi︠
%html
<a rel="license" href="http://creativecommons.org/licenses/by-sa/3.0/us/"><img alt="Creative Commons License" style="border-width:0" src="http://i.creativecommons.org/l/by-sa/3.0/us/88x31.png" /></a><br />This work by <span xmlns:cc="http://creativecommons.org/ns#" property="cc:attributionName">Jason Grout</span> is licensed under a <a rel="license" href="http://creativecommons.org/licenses/by-sa/3.0/us/">Creative Commons Attribution-Share Alike 3.0 United States License</a>.

︡573f61be-bdfa-4f4c-a999-d7dc40d8fe5a︡{"html": "<a rel=\"license\" href=\"http://creativecommons.org/licenses/by-sa/3.0/us/\"><img alt=\"Creative Commons License\" style=\"border-width:0\" src=\"http://i.creativecommons.org/l/by-sa/3.0/us/88x31.png\" /></a><br />This work by <span xmlns:cc=\"http://creativecommons.org/ns#\" property=\"cc:attributionName\">Jason Grout</span> is licensed under a <a rel=\"license\" href=\"http://creativecommons.org/licenses/by-sa/3.0/us/\">Creative Commons Attribution-Share Alike 3.0 United States License</a>."}︡
